/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.awt;

import net.hizlab.kagetaka.util.Environment;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.SystemColor;
import java.awt.Window;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Hashtable;

/**
 * ̾Υġåפ󶡤ޤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.10 $
 */
public class ToolTip
{
	private static final Font DEFAULT_FONT = new Font("Dialog" , Font.PLAIN, 12);
	private static final int  MARGIN = 3;
	private static final int  SPACE  = 15;
	
	private volatile long delay;
	private volatile long term;
	private Font      font;
	private Color     colorFore;
	private Color     colorBack;
	private Color     colorEdge;
	private Hashtable attached  = new Hashtable();
	private ToolTipWindow       window;
	private ToolTipThread       timer;
	private FocusListener       focusListener;
	private KeyListener         keyListener;
	private MouseListener       mouseListener;
	private MouseMotionListener mouseMotionListener;
	
	
	/**
	 * ꤵ줿ٱɽġåפޤ
	 * 
	 * @param     delay ٱʥߥá
	 * @param     term  ɽ֡ʥߥá
	 */
	public ToolTip(long delay, long term)
	{
		this.delay = delay;
		this.term  = term;
		
		timer = new ToolTipThread("Kankana-ToolTip");
		timer.start();
		
		mouseListener = new MouseAdapter()
		{
			/** ޥä */
			public void mouseEntered(MouseEvent e)
			{
				timer.start(e);
			}
			
			/** ޥФ */
			public void mouseExited(MouseEvent e)
			{
				timer.cancel();
			}
			
			/** ޥ줿 */
			public void mousePressed(MouseEvent e)
			{
				timer.reset(e);
			}
			
			/** ޥ줿 */
			public void mouseReleased(MouseEvent e)
			{
				timer.reset(e);
			}
		};
		mouseMotionListener = new MouseMotionAdapter()
		{
			/** ޥư */
			public void mouseMoved(MouseEvent e)
			{
				timer.reset(e);
			}
		};
		focusListener = new FocusAdapter()
		{
			/** ե줿 */
			public void focusLost(FocusEvent e)
			{
				timer.cancel();
			}
		};
		keyListener = new KeyAdapter()
		{
			/** 줿 */
			public void keyPressed(KeyEvent e)
			{
				timer.cancel();
			}
		};
		
		setFont(DEFAULT_FONT);
		setForeground(SystemColor.infoText    );
		setBackground(SystemColor.info        );
		setEdgeColor (SystemColor.windowBorder);
	}
	
	/**
	 * եȤ֤ޤ
	 * 
	 * @return    ե
	 */
	public Font getFont()
	{
		return font;
	}
	
	/**
	 * եȤꤷޤ
	 * 
	 * @param     font ե
	 */
	public synchronized void setFont(Font font)
	{
		this.font = font;
		ToolTipWindow w = window;
		if (w != null)
			w.changeText = true;
	}
	
	/**
	 * ʿ֤ޤ
	 * 
	 * @return    ʿ
	 */
	public Color getForeground()
	{
		return colorFore;
	}
	
	/**
	 * ʿꤷޤ
	 * 
	 * @param     c ʿ
	 */
	public synchronized void setForeground(Color c)
	{
		colorFore = c;
		ToolTipWindow w = window;
		if (w != null)
			w.changeText = true;
	}
	
	/**
	 * طʿ֤ޤ
	 * 
	 * @return    طʿ
	 */
	public Color getBackground()
	{
		return colorBack;
	}
	
	/**
	 * طʿꤷޤ
	 * 
	 * @param     c طʿ
	 */
	public synchronized void setBackground(Color c)
	{
		colorBack = c;
		ToolTipWindow w = window;
		if (w != null)
			w.changeText = true;
	}
	
	/**
	 * ο֤ޤ
	 * 
	 * @return    ο
	 */
	public Color getEdgeColor()
	{
		return colorEdge;
	}
	
	/**
	 * οꤷޤ
	 * 
	 * @param     c ο
	 */
	public synchronized void setEdgeColor(Color c)
	{
		colorEdge = c;
		ToolTipWindow w = window;
		if (w != null)
			w.changeText = true;
	}
	
	/**
	 * ɽåäޤ
	 */
	public void cancel()
	{
		timer.cancel();
	}
	
	/**
	 * ɽٱ֤ޤ
	 * 
	 * @return    ٱʥߥá
	 */
	public long getDelay()
	{
		return delay;
	}
	
	/**
	 * ɽٱ֤ߥäǻꤷޤ
	 * 
	 * @param     delay ٱʥߥá
	 */
	public void setDelay(long delay)
	{
		this.delay = delay;
	}
	
	/**
	 * ɽ֤ޤ
	 * 
	 * @return    ɽ֡ʥߥá
	 */
	public long getTerm()
	{
		return term;
	}
	
	/**
	 * ɽ֤ߥäǻꤷޤ
	 * 
	 * @param     term ɽ֡ʥߥá
	 */
	public void setTerm(long term)
	{
		this.term = term;
	}
	
	/**
	 * ݡͥȤϿޤ
	 * 
	 * @param     c Ͽ륳ݡͥ
	 * @param     tip ΥݡͥȤɽå
	 */
	public synchronized void attachComponent(Component c, String tip)
	{
		if (attached.containsKey(c))
			return;
		
		attached.put(c, tip);
		c.addKeyListener        (keyListener        );
		c.addMouseListener      (mouseListener      );
		c.addMouseMotionListener(mouseMotionListener);
		if (Environment.javaVersion >= 104)
			c.addFocusListener(focusListener);
		
		if (c instanceof Container) {
			Component[] cs = ((Container)c).getComponents();
			for (int i = 0; i < cs.length; i++)
				attachComponent(cs[i], tip);
		}
	}
	
	/**
	 * ݡͥȤޤ
	 * 
	 * @param     c 륳ݡͥ
	 */
	public synchronized void detachComponent(Component c)
	{
		if (!attached.containsKey(c))
			return;
		
		attached.remove(c);
		c.removeKeyListener        (keyListener        );
		c.removeMouseListener      (mouseListener      );
		c.removeMouseMotionListener(mouseMotionListener);
		if (Environment.javaVersion >= 104)
			c.removeFocusListener(focusListener);
		
		if (c instanceof Container) {
			Component[] cs = ((Container)c).getComponents();
			for (int i = 0; i < cs.length; i++)
				detachComponent(cs[i]);
		}
	}
	
	/**
	 * ˴Ƥ٤ƤΥ꥽ޤ
	 */
	public void dispose()
	{
		timer.interrupt();
		ToolTipWindow w = window;
		if (w != null)
			w.dispose();
	}
	
//### ToolTipWindow
	/** ɥ */
	private class ToolTipWindow
		extends Window
		implements ImageCreator
	{
		private String  text;
		private boolean changeText = true;
		private Object  lock       = new Object();
		
		private int width;
		private int height;
		
		private FontMetrics  fm        = null;
		private int          baseFull  = 0;
		private int          baseHalf  = 0;
		
		private Image     offscreen;
		private Graphics  offscreenGraphics;
		
		/** 󥹥󥹤 */
		private ToolTipWindow()
		{
			super(HiddenFrame.getInstance());
			pack();
		}
		
		/** ɽʸ */
		private void setText(String text)
		{
			synchronized (lock) {
				if (this.text != null && this.text.compareTo(text) == 0)
					return;
				
				this.text = text;
				changeText = true;
				calculate();
			}
		}
		
		/** 礭Ʒ׻ */
		private void calculate()
		{
			FontData fontData = FontData.getInstance(this, font);
			
			this.fm       = fontData.getFontMetrics();
			this.baseFull = fontData.getFullBase();
			this.baseHalf = fontData.getHalfBase();
			
			int width  = fm.stringWidth(text) + MARGIN * 2 + 2;
			int height = Math.max(fontData.getFullSize().height, fontData.getHalfHeight()) + MARGIN * 2 + 2;
			
			if (this.width != width || this.height != height) {
				if (offscreen != null) {
					offscreenGraphics.dispose();
					offscreen.flush();
				}
				
				offscreen         = createImage(width, height);
				offscreenGraphics = offscreen.getGraphics();
			}
			
			setSize(width, height);
			this.width  = width;
			this.height = height;
		}
		
		/**  */
		public void update(Graphics g)
		{
			paint(g);
		}
		
		/**  */
		public void paint(Graphics g)
		{
			Image image;
			int   width;
			int   height;
			
			synchronized (lock) {
				image  = this.offscreen;
				width  = this.width;
				height = this.height;
				
				if (changeText) {
					// 
					
					offscreenGraphics.setColor  (colorBack);
					offscreenGraphics.fillRect  (1, 1, width - 3, height - 3);
					offscreenGraphics.setColor  (colorEdge);
					offscreenGraphics.drawRect  (0, 0, width - 1, height - 1);
					offscreenGraphics.setColor  (colorFore);
					int x = MARGIN + 1;
					int y = MARGIN + 1;
					char[] chars  = text.toCharArray();
					int    length = chars.length;
					char c;
					for (int i = 0; i < length; i++) {
						c = chars[i];
						offscreenGraphics.drawChars(chars, i, 1, x, y + ((c <= 0xFF) || (0xFF61 <= c && c <= 0xFF9F) ? baseHalf : baseFull));
						x += fm.charWidth(chars[i]);
					}
					
					changeText = false;
				}
			}
			
			g.drawImage(image, 0, 0, width, height, this);
		}
		
		/** ˴ */
		public void dispose()
		{
			if (offscreen != null) {
				offscreenGraphics.dispose();
				offscreen.flush();
				
				offscreenGraphics = null;
				offscreen = null;
			}
			super.dispose();
		}
		
		/** ɽ */
		public void show()
		{
			toFront();
			super.show();
		}
	}
	
//### ToolTipThread
	/** ɽԤΥå */
	private class ToolTipThread
		extends Thread
	{
		private Component component;
		private boolean   reset;
		private int x, y;
		
		/** 󥹥󥹤 */
		private ToolTipThread(String name)
		{
			super(name);
		}
		
		/** ¹Ԥ */
		public void run()
		{
			try {
				synchronized (this) {
					for (;;) {
						wait();
						
						while (component != null && reset) {
							reset = false;
							wait(delay);
							if (reset)
								continue;
							
							if (component != null) {
								show((String)attached.get(component));
								wait(term);
								hide();
							}
						}
					}
				}
			} catch (InterruptedException e) {
			}
		}
		
		/**  */
		private synchronized void start(MouseEvent e)
		{
			component = e.getComponent();
			reset     = true;
			calculate(e);
			notify();
		}
		
		/** λ */
		private synchronized void cancel()
		{
			component = null;
			reset     = true;
			hide  ();
			notify();
		}
		
		/** ľ */
		private synchronized void reset(MouseEvent e)
		{
			component = e.getComponent();
			reset     = true;
			hide();
			calculate(e);
			notify();
		}
		
		/** ֤κƷ׻ */
		private void calculate(MouseEvent e)
		{
			Point p = e.getComponent().getLocationOnScreen();
			x = p.x + e.getX();
			y = p.y + e.getY();
		}
		
		/** ɽ */
		private synchronized void show(String tip)
		{
			if (tip == null)
				return;
			
			if (window == null)
				window = new ToolTipWindow();
			
			window.setText(tip);
			
			int x = this.x + SPACE;
			int y = this.y + SPACE;
			Dimension screenSize = window.getToolkit().getScreenSize();
			Dimension windowSize = window.getSize();
			screenSize.width  -= SPACE * 3;
			screenSize.height -= SPACE * 3;
			
			if (x + windowSize.width  > screenSize.width )
				x = screenSize.width  - windowSize.width;
			if (y + windowSize.height > screenSize.height)
				y = this.y - SPACE - windowSize.height;
			
			window.setLocation(x, y);
			window.show();
		}
		
		/**  */
		private synchronized void hide()
		{
			if (window == null || !window.isVisible())
				return;
			
			window.setVisible(false);
		}
	}
}
