/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.rendering;

import net.hizlab.kagetaka.awt.ImageCreator;
import net.hizlab.kagetaka.awt.InnerToolTip;
import net.hizlab.kagetaka.awt.image.OffscreenObserver;
import net.hizlab.kagetaka.token.FormItem;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.image.ImageObserver;
import java.util.Vector;

/**
 * 襭ѥ󶡤ޤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.9 $
 */
class Canvas
	extends Container
{
	private static final int TIP_BOTTOM = 22;
	
	/** @serial 륳ƥ */
	private HawkContext hawkContext;
	
	/** @serial ե꡼ */
	private Object    offscreenLock     = new Object();
	/** @serial ե꡼ */
	private Image     offscreen         = null;
	/** @serial ե꡼礭 */
	private Dimension offscreenSize     = null;
	/** @serial ե꡼󥰥եå */
	private Graphics  offscreenGraphics = null;
	/** @serial ե꡼Ѥߤ x  */
	private int       offscreenPointer  = 0;
	/** @serial ե꡼μ褹ǥå */
	private int       offscreenIndex    = 0;
	/** @serial ᡼å */
	private Object    tilesLock         = new Object();
	/** @serial ᡼ */
	private Tile[]    tiles             = new Tile[10];
	/** @serial ᡼ο */
	private int       tilesCount        = 0;
	/** @serial ᡼ */
	private int       tilesWidth        = 0;
	/** @serial 褹뱦ΰ */
	private Point     position          = new Point(0, 0);
	/** @serial եॢƥ */
	private Vector    formItems         = new Vector();
	/** @serial եॢƥäƤ뤫ɤ */
	private boolean   hasFormItems      = false;
	/** @serial ġå */
	private ToolTip   toolTip;
	
	/** @serial طʿ */
	private Color     backColor         = null;
	/** @serial طʲ */
	private Image     backImage         = null;
	/** @serial طʲ */
	private Dimension backImageSize     = null;
	
	/**
	 * 襭ѥޤ
	 * 
	 * @param     hawkContext 륳ƥ
	 */
	Canvas(HawkContext hawkContext)
	{
		this.hawkContext = hawkContext;
		setLayout(null);
		
		// ꥹʤϿ
		addComponentListener(
			new ComponentAdapter()
			{
				/** ѹ */
				public void componentResized(ComponentEvent e)
				{
					synchronized (offscreenLock) {
						if (offscreen != null) {
							if (offscreenSize.equals(getSize()))
								return;
							
							clearOffscreen();
							repaint();
						}
					}
				}
				
				/** 줿Ȥ */
				public void componentHidden(ComponentEvent e)
				{
					clearOffscreen();
				}
			}
		);
		
		setBackground(backColor = Document.DEFAULT_BACKGROUND_COLOR);
	}
	
	/**
	 * ꡼󥤥᡼ɲäޤ
	 * 
	 * @param     image  ꡼󥤥᡼
	 * @param     width  ꡼󥤥᡼
	 * @param     height ꡼󥤥᡼ι⤵
	 */
	void addImage(Image image, int width, int height)
	{
		synchronized (tilesLock) {
			// ĥ
			if (tilesCount == tiles.length) {
				int    newSize  = (int)(tilesCount * (tilesCount > 300 ? 1.5 : 2));
				Tile[] newTiles = new Tile[newSize];
				System.arraycopy(tiles, 0, newTiles, 0, tilesCount);
				tiles = newTiles;
			}
			
			Tile tile = tiles[tilesCount];
			if (tile == null) {
				tile = new Tile(image, width, height);
				tiles[tilesCount] = tile;
				if (tilesCount > 0)
					tiles[tilesCount - 1].next = tile;
			} else {
				tile.reset(image, width, height);
			}
			tilesWidth += width;
			tilesCount++;
		}
		
		synchronized (offscreenLock) {
			if (isShowing() &&
			    (offscreen == null || offscreenPointer > 0))
				repaint();
		}
	}
	
	/**
	 * եॢƥɲäޤ
	 * 
	 * @param     item եॢƥ
	 */
	void addFormItem(FormItem item)
	{
		hasFormItems = true;
		formItems.addElement(item);
		
		int posX, posY;
		synchronized (position) {
			posX = position.x;
			posY = position.y;
		}
		
		synchronized (tilesLock) {
			if (offscreen == null)
				return;
			
			Point     pos  = item.getPosition();
			Dimension size = item.getSize    ();
			if (posX <= pos.x + size.width  &&
			    posY <= pos.y + size.height &&
			    posX + offscreenSize.width  >= pos.x &&
			    posY + offscreenSize.height >= pos.y) {
				Component c = item.getComponent();
				add(c);
				c.setBounds(offscreenSize.width - pos.x + posX - size.width,
				            pos.y - posY,
				            size.width,
				            size.height);
			}
		}
	}
	
	/**
	 * ѥϰϤꤷƺɽޤ
	 * ɸϱ夫εΥǤ
	 * 
	 * @param     x ѥα顢ɽ֤αüεΥ
	 * @param     y ѥξ夫顢ɽ֤ξüεΥ
	 * @param     width  ɽϰϤ
	 * @param     height ɽϰϤι⤵
	 */
	void repaintImage(int x, int y, int width, int height)
	{
//System.out.println(x+","+y+","+width+","+height);
		Tile[] tiles;
		int    tilesCount;
		synchronized (tilesLock) {
			tiles      = this.tiles;
			tilesCount = this.tilesCount;
		}
		
		int posX, posY;
		synchronized (position) {
			posX = position.x;
			posY = position.y;
		}
		
		synchronized (offscreenLock) {
			if (offscreen == null) {
				repaint();
				return;
			}
			
			// ϰϡʺ夫εΥ
			int base   = offscreenSize.width + posX;
			int right  = base - x;
			int left   = right - width;
			int top    = y - posY;
			int bottom = top + height;
			int index;
			
			if (right  < 0 || offscreenSize.width  <= left ||
			    bottom < 0 || offscreenSize.height <= top  ||
			    (index = getTileIndex(tiles, tilesCount, x, 0)) == -1)
				return;
			
			int x1;
			int x2;
			int y2 = y + height;
			
			Tile              tile  = tiles[index];
			boolean           first = true;
			OffscreenObserver oo    = null;
			boolean           sync  = false;
			
			// 岼ˤϤ߽Фʬΰ֤Ĵ
			if (top < 0) {
				y   -= top;
				top =  0;
			}
			if (offscreenSize.height < bottom) {
				int diff = bottom - offscreenSize.height;
				y2     -= diff;
				bottom -= diff;
			}
			
			do {
				if (!sync) {
					oo   = new OffscreenObserver(this, offscreenGraphics);
					oo.y = top;
				}
				
				x1 = base - tile.left;
				
				if (x1 <= left) {
					// ǽäƤ
					oo.x = left;
					x1 = left - x1;
					
					if (first)
						// ϰϤҤȤĤΥˤäݤäƤ
						x2 = x1 + width;
					else
						// Υ뤫ĤʤäƤ
						x2 = tile.width;
					sync = offscreenGraphics.drawImage(tile.image,
					                                   oo.x, top, right, bottom,
					                                   x1, y, x2, y2,
					                                   oo);
					break;
				} else {
					// ޤ³Ƥ
					oo.x = x1;
					
					if (first) {
						// ΥǻϤޤꡢΥ³Ƥ
						x2 = right - x1;
						first = false;
					} else
						// Υ뤫ϤޤꡢΥ³Ƥ
						x2 = tile.width;
					sync = offscreenGraphics.drawImage(tile.image,
					                                   oo.x, top, right, bottom,
					                                   0, y, x2, y2,
					                                   oo);
					right = x1;
				}
				tile = tile.next;
			} while (++index < tilesCount);
			
			// ƥ
			paintItem();
			
			Graphics g = getGraphics();
			if (g != null)
				try {
					g.drawImage(offscreen, 0, 0, this);
				} finally {
					g.dispose();
				}
			else
				repaint();
		}
	}
	
	/**
	 * 򥯥ꥢޤ
	 * 
	 * @param     color طʿ
	 * @param     image طʲ
	 */
	void setupScreen(Color color, Image image)
	{
		synchronized (position) {
			position.x = 0;
			position.y = 0;
		}
		
		synchronized (offscreenLock) {
			if (offscreen != null)
				offscreenPointer = offscreenSize.width;
			offscreenIndex = 0;
		}
		
		synchronized (tilesLock) {
			for (int i = 0; i < tilesCount; i++) {
				tiles[i].image.flush();
				tiles[i].next = null;
				tiles[i] = null;
			}
			
			tilesCount = 0;
			tilesWidth = 0;
			
			formItems.removeAllElements();
			hasFormItems = false;
			removeAll();
		}
		
		backColor = color;
		backImage = image;
		super.setBackground(color);
		if (image != null)
			backImageSize = new Dimension(image.getWidth(this), image.getHeight(this));
	}
	
	/**
	 * 褹륤᡼ΰ֤ѹޤ
	 * 
	 * @param     x 夫εΥ
	 * @param     y 夫εΥ
	 */
	void movePosition(int x, int y)
	{
		synchronized (position) {
			position.x = x;
			position.y = y;
		}
		
		synchronized (offscreenLock) {
			if (offscreen == null)
				return;
			
			// 褵褦ˡݥ󥿤᤹
			offscreenPointer = offscreenSize.width;
			offscreenIndex   = 0;
			
			// եॢƥ֤
			if (hasFormItems)
				allocateFormItems(x, y);
		}
		
		Graphics g = getGraphics();
		if (g != null)
			try {
				paint(g);
			} finally {
				g.dispose();
			}
		else
			repaint();
	}
	
	/**
	 * ġåפɽޤ
	 * 
	 * @param     tip   ʸ
	 * @param     x     X
	 * @param     y     Y
	 * @param     delay ɽޤǤԵ֡ʥߥá
	 * @param     term  ɽ֡ʥߥá
	 */
	void showToolTip(String tip, int x, int y, long delay, long term)
	{
		Dimension size = offscreenSize;
		int posX, posY;
		synchronized (position) {
			posX = position.x;
			posY = position.y;
			
			// ĤǤ...
			if (toolTip == null) {
				toolTip = new ToolTip(hawkContext, delay, term, new Insets(0, 0, TIP_BOTTOM * 2, 0));
			} else {
				toolTip.setDelay(delay);
				toolTip.setTerm (term );
			}
		}
		
		toolTip.showTip(tip, x, y + TIP_BOTTOM, size.width, size.height, posX, posY);
	}
	
	/**
	 * ġåפäޤ
	 */
	void clearToolTip()
	{
		if (toolTip != null)
			toolTip.cancel();
	}
	
	/**
	 * ꥽ޤ
	 */
	void cleanup()
	{
		if (toolTip != null) {
			toolTip.dispose();
			toolTip = null;
		}
		setupScreen(null, null);
		clearOffscreen();
	}
	
//### Override
	/**
	 * 褷ޤ
	 * 
	 * @param     g Graphics
	 */
	public void update(Graphics g)
	{
		paint(g);
	}
	
	/**
	 * 褷ޤ
	 * 
	 * @param     g Graphics
	 */
	public void paint(Graphics g)
	{
		int posX, posY;
		synchronized (position) {
			posX = position.x;
			posY = position.y;
		}
		
		Tile[] tiles;
		int    tilesCount;
		synchronized (tilesLock) {
			tiles      = this.tiles;
			tilesCount = this.tilesCount;
		}
		
		synchronized (offscreenLock) {
			// ե꡼Хåե̵Ϻ
			if (offscreen == null) {
				Dimension size = getSize();
				
				offscreen         = createImage(size.width, size.height);
				offscreenSize     = size;
				offscreenGraphics = offscreen.getGraphics();
				offscreenPointer  = size.width;
				offscreenIndex    = 0;
				
				// եॢƥ֤
				if (hasFormItems)
					allocateFormItems(posX, posY);
			}
			
			// 襤᡼򥳥ԡ
			int  index = offscreenIndex;
			if (offscreenPointer > 0 &&           // ޤ̤褬ĤäƤ
			    offscreenIndex < tilesCount &&    // ̤Υ뤬ĤäƤ
			    (tiles[index].left > posX ||      // Υ뤬̾ˤ
			     (index = getTileIndex(tiles, tilesCount, posX, offscreenIndex)) >= 0)) {
				Tile tile     = tiles[index];
				int  osHeight = offscreenSize.height;
				OffscreenObserver oo   = null;
				boolean           sync = false;
				int x2, y2;
				
				// 褹륿αüΰ֤Ĵ
				offscreenPointer = offscreenSize.width - (tile.right - posX);
				do {
					if (!sync)
						oo = new OffscreenObserver(this, offscreenGraphics);
					
					x2   = offscreenPointer;
					y2   = Math.min(tile.height - posY, osHeight);
					oo.x = (offscreenPointer -= tile.width);
					sync = offscreenGraphics.drawImage(tile.image,
					                                   oo.x, 0, x2, y2,
					                                   0, posY,
					                                   tile.width, Math.min(tile.height, osHeight + posY),
					                                   oo);
					// ¦;򤬤硢طʤԤ
					if (y2 < osHeight)
						drawBackground(oo.x, y2, x2, osHeight, posX, posY);
					
					index++;
					tile = tile.next;
				} while (index < tilesCount && offscreenPointer > 0);
				offscreenIndex = index;
				
				// ¦;򤬤硢طʤԤ
				if (offscreenPointer > 0)
					drawBackground(0, 0, offscreenPointer, offscreenSize.height, posX, posY);
			}
			
			// ƥ
			paintItem();
			
			g.drawImage(offscreen, 0, 0, this);
		}
	}
	
	/**
	 * ե뤳Ȥ뤫֤ޤ
	 * ŪˤΥڥϡեܡɤˤǽʤΤǡ
	 *  <code>true</code> ֤ޤ
	 * 
	 * @return    ե뤳ȤǤ뤫
	 */
	public boolean isFocusTraversable()
	{
		return true;
	}
	
	/**
	 * ΥѥΥѥ᡼ʸ֤ޤ
	 * 
	 * @return    ѥ᡼ʸ
	 */
	protected String paramString()
	{
		String str = super.paramString();
		return str;
	}
	
//### private
	/** ե꡼Хåե򥯥ꥢ */
	private void clearOffscreen()
	{
		synchronized (offscreenLock) {
			if (offscreen != null) {
				offscreenGraphics.dispose();
				offscreen        .flush  ();
			}
			
			offscreen         = null;
			offscreenSize     = null;
			offscreenGraphics = null;
			offscreenPointer  = 0;
			offscreenIndex    = 0;
		}
	}
	
	/** ֤Υ֤֤ */
	private int getTileIndex(Tile[] tiles, int tilesCount, int x, int start)
	{
		if (tiles[tilesCount - 1].left <= x)
			return -1;
		
		int end    = tilesCount - 1;
		int middle;
		int right;
		
		while (start + 1 < end) {
			middle = (end - start) / 2 + start;
			if ((right = tiles[middle].right) == x)
				return middle;
			
			if (right < x)
				start = middle;
			else
				end   = middle;
		}
		
		return start;
	}
	
	/** ϰϤطʿطʲԤ - ºݤ x,y ǻ( 0,0) offscreenLock */
	private void drawBackground(int x1, int y1, int x2, int y2, int posX, int posY)
	{
//System.out.println("a="+x1+","+y1+","+x2+","+y2);
		
		if (backImage == null) {
			offscreenGraphics.setColor(backColor);
			offscreenGraphics.fillRect(x1, y1, x2 - x1, y2 - y1);
		} else {
			int sx1 = 0, sy1 = 0, sx2 = 0, sy2 = 0;
			int dx1 = 0, dy1 = 0, dx2 = 0, dy2 = 0;
			
			dx1 = ((posX + (offscreenSize.width - x2)) % backImageSize.width) + x2;
			int offsetY = ((posY + y1) % backImageSize.width);
			do {
				// x Ĵ
				sx1 =  0;
				sx2 =  backImageSize.width;
				dx2 =  dx1;
				dx1 -= backImageSize.width;
				// ¦ˤϤ߽ФƤʤå
				if (dx2 > x2) {
					sx2 -= dx2 - x2;
					dx2 =  x2;
				}
				// ¦ˤϤ߽ФƤʤå
				if (dx1 < x1) {
					sx1 = x1 - dx1;
					dx1 = x1;
				}
				// y Ĵ
				sy2 = backImageSize.height;
				dy2 = y1 - offsetY;
//System.out.println("d="+dx1+","+dx2+","+sx1+","+sx2);
				do {
					sy1 =  0;
					dy1 =  dy2;
					dy2 += backImageSize.height;
					// ¦ˤϤ߽ФƤʤå
					if (dy1 < y1) {
						sy1 = y1 - dy1;
						dy1 = y1;
					}
					// ¦ˤϤ߽ФƤʤå
					if (dy2 > y2) {
						sy2 -= dy2 - y2;
						dy2 =  y2;
					}
//System.out.println(dx1+","+dy1+","+dx2+","+dy2+","+sx1+","+sy1+","+sx2+","+sy2);
					offscreenGraphics.drawImage(backImage, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, this);
				} while (dy2 < y2);
			} while (dx1 > x1);
		}
	}
	
	/** եॢƥෲ֡ offscreenLock */
	private void allocateFormItems(int posX, int posY)
	{
		for (int i = formItems.size() - 1; i >= 0; i--)
			allocateFormItem((FormItem)formItems.elementAt(i), posX, posY);
	}
	
	/** եॢƥ֡ offscreenLock */
	private void allocateFormItem(FormItem item, int posX, int posY)
	{
		Point     pos  = item.getPosition ();
		Dimension size = item.getSize     ();
		Component c    = item.getComponent();
		
		if (posX <= pos.x + size.width  &&
		    posY <= pos.y + size.height &&
		    posX + offscreenSize.width  >= pos.x &&
		    posY + offscreenSize.height >= pos.y) {
			if (c.getParent() == null)
				add(c);
			c.setBounds(offscreenSize.width - pos.x + posX - size.width,
			            pos.y - posY,
			            size.width,
			            size.height);
		} else {
			if (c.getParent() != null)
				remove(c);
		}
	}
	
	/** ƥʥեࡢġåסˤ offscreenLock */
	private void paintItem() {
		// ե
		if (hasFormItems) {
			Component cs[] = getComponents();
			
			Graphics  g;
			Component c;
			Rectangle r;
			for (int i = cs.length - 1; i >= 0; i--) {
				if ((c = cs[i]) != null) {
					r = c.getBounds();
					g = offscreenGraphics.create(r.x, r.y, r.width, r.height);
					
					try {
						c.paintAll(g);
					} finally {
						g.dispose();
					}
				}
			}
		}
		
		// ġåפ
		ToolTip toolTip = this.toolTip;
		if (toolTip != null)
			toolTip.paint(offscreenGraphics);
	}
	
//### Tile
	/** ᡼ */
	private class Tile
	{
		private Image image;
		private int   width;
		private int   height;
		private int   right;
		private int   left;
		private Tile  next;
		
		/** 󥹥󥹤 */
		private Tile(Image image, int width, int height)
		{
			reset(image, width, height);
		}
		
		/**  */
		private void reset(Image image, int width, int height)
		{
			this.image  = image;
			this.width  = width;
			this.height = height;
			this.right  = tilesWidth;
			this.left   = right + width;
		}
	}
	
//### ToolTip
	/** ġå */
	private class ToolTip
		extends InnerToolTip
	{
		private int left;
		private int top;
		
		/** 󥹥󥹤 */
		private ToolTip(ImageCreator creator, long delay, long term, Insets fix)
		{
			super(creator, delay, term, fix);
		}
		
		/** ɽ */
		public void showTip(String value, int x, int y, int width, int height, int posX, int posY)
		{
			super.showTip(value, x, y, width, height);
			this.left = width + posX;
			this.top  = posY;
		}
		
		/** ɽ */
		public void repaint(int x, int y, int width, int height)
		{
			repaintImage(left - x - width, top + y, width, height);
		}
	}
}
