/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.viewer;

import net.hizlab.kagetaka.Reporter;
import net.hizlab.kagetaka.Resource;
import net.hizlab.kagetaka.awt.MessageBox;
import net.hizlab.kagetaka.cookie.Cookie;
import net.hizlab.kagetaka.cookie.CookieManager;
import net.hizlab.kagetaka.option.ViewerOption;
import net.hizlab.kagetaka.protocol.ProxySupported;
import net.hizlab.kagetaka.rendering.PostData;
import net.hizlab.kagetaka.rendering.Request;

import java.awt.Frame;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.ConnectException;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.Hashtable;

/**
 * ץȥ³򥵥ݡȤ륯饹Ǥ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.14 $
 */
class Connector
{
	private static final String RESOURCE = "net.hizlab.kagetaka.viewer.Resource";
	
	private static Hashtable authCache  = new Hashtable();
	private static String    proxyAuthorization;
	
	private ViewerContext context;
	private ViewerOption  option;
	private Reporter      reporter;
	private Frame         owner;
	private CookieManager cookieManager;
	
	/**
	 * ͥΥ󥹥󥹤ޤ
	 */
	Connector(ViewerContext context, ViewerOption option)
	{
		this.context       = context;
		this.reporter      = context.getReporter();
		this.option        = option;
		this.owner         = owner;
		this.cookieManager = option.getCookieManager();
	}
	
	/** ʡ */
	void setOwner(Frame owner)
	{
		this.owner = owner;
	}
	
	/** ꤵ줿 URL Υ֤ͥ */
	URLConnection getURLConnection(Request request)
		throws IOException, InterruptedException
	{
		URLConnection     connection     = null;
		HttpURLConnection httpConnection = null;
		URL               url            = request.url;
		URL               proxyUrl       = null;
		boolean           debug          = (reporter != null && reporter.getLevel() <= Reporter.DEBUG);
		
		try {
			if (reporter != null)
				reporter.report(Reporter.INFO, getMessage("report.connect.open", new String[]{url.toString()}), 0, 0);
			
			URL                baseUrl              = url;
			URL                connectUrl           = url;
			PostData           pd                   = request.postData;
			boolean            isNewUrl             = true;
			boolean            isNewProxyAuth       = false;
			AuthenticationInfo proxyAuthentication  = null;
			boolean            isNewServerAuth      = false;
			AuthenticationInfo serverAuthentication = null;
			
			CONNECT:
			for (;;) {
				// ǧھ
				if (isNewUrl) {
					// ǧھ
					serverAuthentication = AuthenticationInfo.get(AuthenticationInfo.SERVER, url);
					
					// ץǧھ
					URL newProxyUrl = option.getProxyURL(url);
					if (newProxyUrl != proxyUrl) {
						proxyUrl = newProxyUrl;
						if (proxyUrl != null)
							proxyAuthentication = AuthenticationInfo.get(AuthenticationInfo.PROXY, proxyUrl);
						else
							proxyAuthentication = null;
					}
					
					// URL ץ
					if (proxyUrl != null)
						connectUrl = new URL(url.getProtocol(), proxyUrl.getHost(), proxyUrl.getPort(), url.toString());
					else
						connectUrl = url;
				}
				
				// Keep-Alive ڤ礬Τǡ٤ƻԤ뤿Υ롼
				boolean firstTry = true;
				RETRY:
				for (;;) {
					connection = connectUrl.openConnection();
					if (connection instanceof HttpURLConnection)
						httpConnection = (HttpURLConnection)connection;
					else
						httpConnection = null;
					
					// ǥեȤ
					connection.setUseCaches(firstTry);
					
					// ꥯȥإå
					setHTTPHeader(url, connection, request, proxyAuthentication, serverAuthentication);
					
					// POST ǡ³
					if (pd != null && httpConnection != null)
						pd.send(httpConnection);
					
					// ³
					try {
						connection.connect();
						
						// ǳμ¤³
						if (httpConnection != null)
							httpConnection.getResponseCode();
						else
							connection.getInputStream();
					} catch (SocketException e) {
						if (!firstTry ||
						    e.getClass().getName().compareTo("java.net.SocketException") != 0)
							throw e;
						firstTry = false;
						continue RETRY;
					} finally {
						if (Thread.interrupted())
							throw new InterruptedException("connect");
					}
					break RETRY;
				}
				isNewUrl = false;
				
				// 쥹ݥ󥹥إå
				getHTTPHeader(url, connection);
				
				if (debug)
					dumpConnection(connection);
				
				// ʹߤ HTTP ν
				if (httpConnection == null)
					break;
				
				int code = httpConnection.getResponseCode();
				
				if (code == HttpURLConnection.HTTP_UNAUTHORIZED) {
					// ǧڽ
					if ((serverAuthentication = getWWWAuthenticate(url, connection, isNewServerAuth, false)) == null)
						break CONNECT;
					isNewServerAuth = true;
					
					// ɹԤ
					if (reporter != null)
						reporter.report(Reporter.INFO, getMessage("report.connect.reopen", new String[]{url.toString()}), 0, 0);
				} else if (code == HttpURLConnection.HTTP_PROXY_AUTH) {
					// ǧڽ
					if ((proxyAuthentication = getWWWAuthenticate(proxyUrl, connection, isNewProxyAuth, true)) == null)
						break CONNECT;
					isNewProxyAuth = true;
					
					// ɹԤ
					if (reporter != null)
						reporter.report(Reporter.INFO, getMessage("report.connect.reopen", new String[]{url.toString()}), 0, 0);
				} else {
					// ǧھϤ줬ϥåɲ
					if (isNewProxyAuth )
						AuthenticationInfo.put(AuthenticationInfo.PROXY , proxyAuthentication , proxyUrl);
					if (isNewServerAuth)
						AuthenticationInfo.put(AuthenticationInfo.SERVER, serverAuthentication, url     );
					
					switch (code) {
					case HttpURLConnection.HTTP_MULT_CHOICE:         // 300
					case HttpURLConnection.HTTP_MOVED_PERM:          // 301
					case 307:                                        // 307 Temporary Redirect
						break;
					case HttpURLConnection.HTTP_MOVED_TEMP:          // 302
						if (pd != null && !option.getPropertyBoolean(ViewerOption.KEY_HTTP_STRICT, false))
							pd = null;
						break;
					case HttpURLConnection.HTTP_SEE_OTHER :          // 303
						if (pd != null)
							pd = null;
						break;
					default:
						break CONNECT;
					}
					
					// 쥯Ƚ
					if ((url = getRedirectURL(url, connection, pd, baseUrl)) == null)
						break CONNECT;
					
					// ǧھ
					isNewUrl             = true;
					isNewProxyAuth       = false;
					isNewServerAuth      = false;
					serverAuthentication = null;
					
					if (reporter != null)
						reporter.report(Reporter.INFO, getMessage("report.connect.redirect", new String[]{url.toString()}), 0, 0);
				}
				
				httpConnection.disconnect();
			}
			
			return ((proxyUrl == null) || (connection instanceof ProxySupported)
			        ? connection
			        : (httpConnection != null
			           ? (URLConnection)new HttpURLConnectionWrapper(url, httpConnection)
			           : (URLConnection)new URLConnectionWrapper    (url, connection    )));
		} catch (IOException e) {
			// ץ³顼ʤ⤷ʤ˾
			if (proxyUrl != null &&
			    (e instanceof UnknownHostException ||
			     e instanceof ConnectException     ||
			     e instanceof SocketException      )) {
				String option = (e instanceof UnknownHostException ? "notfound" : "notconnect");
				MessageBox.show(owner,
				                getMessage("message.proxy." + option + ".text" , new String[]{proxyUrl.getHost()}),
				                getMessage("message.proxy." + option + ".title", null),
				                MessageBox.BUTTON_OK | MessageBox.ICON_EXCLAMATION);
				return null;
			}
			
			throw e;
		}
	}
	
	/** å */
	void setCookie(String value, URL url)
	{
		Cookie cookie = cookieManager.createCookie(value, url);
		if (cookie == null)
			return;
		
		boolean strage = (cookie.getExpires() >= 0);
		String  accept;
		if (strage)
			accept = option.getPropertyString(ViewerOption.KEY_COOKIE_ACCEPT_STRAGE );
		else
			accept = option.getPropertyString(ViewerOption.KEY_COOKIE_ACCEPT_SESSION);
		
		if (accept == null || accept.compareTo(CookieManager.ASC) == 0) {
			if (MessageBox.show(owner,
			                    getMessage("message.cookie.text" ,
			                               new Object[]{cookie.getDomain(),
			                                            cookie.getName  (),
			                                            cookie.getValue (),
			                                            cookie.getDomain(),
			                                            cookie.getPath  (),
			                                            new Integer((strage ? 1 : 0)),
			                                            (strage ? new Date(cookie.getExpires()) : null)}),
			                    getMessage("message.cookie.title", null),
			                    MessageBox.BUTTON_YESNO | MessageBox.ICON_QUESTION) == MessageBox.RESULT_YES)
				accept = CookieManager.YES;
			else
				accept = CookieManager.NO;
		}
		
		if (accept.compareTo(CookieManager.YES) == 0)
			cookieManager.setCookie(cookie);
	}
	
	/** HTTP ꥯȥإå */
	private void setHTTPHeader(URL url, URLConnection connection, Request request,
	                           AuthenticationInfo proxyAuthentication,
	                           AuthenticationInfo serverAuthentication)
		throws IOException
	{
		connection.setRequestProperty("Host"           , url.getHost() + (url.getPort() > 0 ? ":" + url.getPort() : ""));
		connection.setRequestProperty("User-Agent"     , option.getPropertyString(ViewerOption.KEY_HTTP_USERAGENT     ));
		connection.setRequestProperty("Accept"         , option.getPropertyString(ViewerOption.KEY_HTTP_ACCEPT        ));
		connection.setRequestProperty("Accept-Language", option.getPropertyString(ViewerOption.KEY_HTTP_ACCEPTLANGUAGE));
		connection.setRequestProperty("Accept-Encoding", option.getPropertyString(ViewerOption.KEY_HTTP_ACCEPTENCODING));
		connection.setRequestProperty("Accept-Charset" , option.getPropertyString(ViewerOption.KEY_HTTP_ACCEPTCHARSET ));
		
		int referer = option.getPropertyInteger(ViewerOption.KEY_HTTP_REFERER, HawkViewer.REFERER_SAMEHOST);
		if (request.referer != null && referer != HawkViewer.REFERER_NONE &&
		    (referer == HawkViewer.REFERER_ALL ||
		     isSamePath(request.referer, url, referer)))
			connection.setRequestProperty("Referer", request.referer.toExternalForm());
		
		String cookie = cookieManager.getCookie(url);
		if (cookie != null)
			connection.setRequestProperty("Cookie", cookie);
		
		if (proxyAuthentication  != null)
			connection.setRequestProperty("Proxy-Authorization", proxyAuthentication .getValue());
		if (serverAuthentication != null)
			connection.setRequestProperty("Authorization"      , serverAuthentication.getValue());
		
		if (request.useCache == Request.CACHE_NONE) {
			connection.setRequestProperty("Cache-Control", "no-cache");
			connection.setRequestProperty("Pragma"       , "no-cache");
		}
	}
	
	/** HTTP ꥯȥإå */
	private void getHTTPHeader(URL url, URLConnection connection)
		throws IOException
	{
		String key;
		int    index = 1;
		
		for (;;) {
			if ((key = connection.getHeaderFieldKey(index)) == null)
				break;
			
			key = key.toLowerCase();
			if (key.compareTo("set-cookie") == 0)
				setCookie(connection.getHeaderField(index), url);
			
			index++;
		}
	}
	
	/** ǧھ桼 */
	private AuthenticationInfo getWWWAuthenticate(URL url, URLConnection connection,
	                                              boolean retry, boolean proxy)
	{
		String realm;
		if (proxy)
			realm = connection.getHeaderField("Proxy-Authenticate");
		else
			realm = connection.getHeaderField("WWW-Authenticate");
		if (realm == null)
			return null;
		
		// ǧھʬ (WWW-Authenticate: Basic realm=dokoka)
		String type = realm;
		int p = realm.indexOf(' ');
		if (p != -1) {
			type  = realm.substring(0, p);
			realm = realm.substring(p + 1);
			if (realm.toLowerCase().startsWith("realm")) {
				for (p = 5; p < realm.length(); p++) {
					if (realm.charAt(p) != ' ') {
						if (realm.charAt(p) == '=')
							realm = realm.substring(p + 1).trim();
						break;
					}
				}
			}
		} else {
			realm = null;
		}
		
		char atype;
		if (type.toLowerCase().compareTo("basic") == 0) {
			atype = AuthenticationInfo.BASIC;
		} else {
			// Basic ʳǧڤϥݡȳ
			MessageBox.show(owner,
			                getMessage("message.unknown_authtype.text" , new String[]{type, url.toString()}),
			                getMessage("message.unknown_authtype.title", null),
			                MessageBox.BUTTON_OK | MessageBox.ICON_EXCLAMATION);
			return null;
		}
		
		if (!retry) {
			// Realm бǧھ򸡺
			AuthenticationInfo auth = AuthenticationInfo.get((proxy ? AuthenticationInfo.PROXY : AuthenticationInfo.SERVER),
			                                                 url, realm, atype);
			if (auth != null)
				return auth;
		}
		
		// 桼˼
		return AuthorizationDialog.show(owner, url.getHost(), realm, atype,
		                                (proxy ? AuthorizationDialog.PROXY : AuthorizationDialog.SERVER));
	}
	
	/** 쥯ȤԤ֤ */
	private URL getRedirectURL(URL url, URLConnection connection, PostData pd, URL baseUrl)
		throws IOException
	{
		String location = connection.getHeaderField("Location");
		if (location == null)
			return null;
		
		try {
			url = new URL(url, location);         //  Location ϥեѥΤϤ
			if (pd != null &&
			    MessageBox.show(owner,
			                    getMessage("message.redirect.text" , new String[]{baseUrl.toString(), location}),
			                    getMessage("message.redirect.title", null),
			                    MessageBox.BUTTON_OKCANCEL | MessageBox.ICON_QUESTION) != MessageBox.BUTTON_OK)
				return null;
		} catch (MalformedURLException e) {
			MessageBox.show(owner,
			                getMessage("message.invalidurl.text" , new String[]{location, e.toString()}),
			                getMessage("message.invalidurl.title", null),
			                MessageBox.BUTTON_OK | MessageBox.ICON_EXCLAMATION);
			return null;
		}
		
		return url;
	}
	
	/** ѥƱå */
	private boolean isSamePath(URL url1, URL url2, int path)
	{
		if (url1.getHost().compareTo(url2.getHost()) != 0)
			return false;
		
		if (url1.getPort() != url2.getPort())
			return false;
		
		String s1 = url1.getFile();
		String s2 = url2.getFile();
		int p;
		
		// ѥ¸ߤʤabout: ʤɡ
		if ((p = s1.indexOf('/')) == -1)
			return true;
		
		while (path-- > 0 && ((p = s1.indexOf('/', p + 1)) != -1))
			;
		if (p == -1)
			p = s1.lastIndexOf('/');
		
		if (s2.length() <= p)
			return false;
		
		for (int i = 0; i <= p; i++) {
			if (s1.charAt(i) != s2.charAt(i))
				return false;
		}
		
		return true;
	}
	
	/** ꥽ʸ */
	private String getMessage(String key, Object[] args)
	{
		return Resource.getMessage(RESOURCE, key, args);
	}
	
	/** ͥ */
	private void dumpConnection(URLConnection connection)
		throws IOException
	{
		if (connection instanceof HttpURLConnection) {
			HttpURLConnection httpConnection = (HttpURLConnection)connection;
			reporter.report(Reporter.DEBUG, "HTTP Response: " + httpConnection.getResponseCode()
			                                            + " " + httpConnection.getResponseMessage(), 0, 0);
		}
		
		int index = 0;
		String key;
		
		while ((key = connection.getHeaderFieldKey(++index)) != null)
			reporter.report(Reporter.DEBUG, "HTTP Response: [" + key + "]:[" + connection.getHeaderField(index) + "]", 0, 0);
	}
	
//### URLConnectionWrapper
	/** URLConnection Υåѡ */
	private class URLConnectionWrapper
		extends URLConnection
	{
		private URLConnection connection;
		
		/** 󥹥󥹤 */
		private URLConnectionWrapper(URL url, URLConnection connection)
		{
			super(url);
			this.connection = connection;
		}
		
		// URLConnection
		public void         connect                   (                            ) throws IOException       {        connection.connect                   (                    ); }
		public URL          getURL                    (                            )                          { return url                                                        ; }
		public int          getContentLength          (                            )                          { return connection.getContentLength          (                    ); }
		public String       getContentType            (                            )                          { return connection.getContentType            (                    ); }
		public String       getContentEncoding        (                            )                          { return connection.getContentEncoding        (                    ); }
		public long         getExpiration             (                            )                          { return connection.getExpiration             (                    ); }
		public long         getDate                   (                            )                          { return connection.getDate                   (                    ); }
		public long         getLastModified           (                            )                          { return connection.getLastModified           (                    ); }
		public String       getHeaderField            (String name                 )                          { return connection.getHeaderField            (name                ); }
		public int          getHeaderFieldInt         (String name, int  Default   )                          { return connection.getHeaderFieldInt         (name, Default       ); }
		public long         getHeaderFieldDate        (String name, long Default   )                          { return connection.getHeaderFieldDate        (name, Default       ); }
		public String       getHeaderFieldKey         (int n                       )                          { return connection.getHeaderFieldKey         (n                   ); }
		public String       getHeaderField            (int n                       )                          { return connection.getHeaderField            (n                   ); }
		public Object       getContent                (                            ) throws IOException       { return connection.getContent                (                    ); }
		public InputStream  getInputStream            (                            ) throws IOException       { return connection.getInputStream            (                    ); }
		public OutputStream getOutputStream           (                            ) throws IOException       { return connection.getOutputStream           (                    ); }
		public String       toString                  (                            )                          { return connection.toString                  (                    ); }
		public void         setDoInput                (boolean doinput             )                          {        connection.setDoInput                (doinput             ); }
		public boolean      getDoInput                (                            )                          { return connection.getDoInput                (                    ); }
		public void         setDoOutput               (boolean dooutput            )                          {        connection.setDoOutput               (dooutput            ); }
		public boolean      getDoOutput               (                            )                          { return connection.getDoOutput               (                    ); }
		public void         setAllowUserInteraction   (boolean allowuserinteraction)                          {        connection.setAllowUserInteraction   (allowuserinteraction); }
		public boolean      getAllowUserInteraction   (                            )                          { return connection.getAllowUserInteraction   (                    ); }
		public void         setUseCaches              (boolean usecaches           )                          {        connection.setUseCaches              (usecaches           ); }
		public boolean      getUseCaches              (                            )                          { return connection.getUseCaches              (                    ); }
		public void         setIfModifiedSince        (long ifmodifiedsince        )                          {        connection.setIfModifiedSince        (ifmodifiedsince     ); }
		public long         getIfModifiedSince        (                            )                          { return connection.getIfModifiedSince        (                    ); }
		public boolean      getDefaultUseCaches       (                            )                          { return connection.getDefaultUseCaches       (                    ); }
		public void         setDefaultUseCaches       (boolean defaultusecaches    )                          {        connection.setDefaultUseCaches       (defaultusecaches    ); }
		public void         setRequestProperty        (String key, String value    )                          {        connection.setRequestProperty        (key, value          ); }
		public String       getRequestProperty        (String key                  )                          { return connection.getRequestProperty        (key                 ); }
//		public void         addRequestProperty        (String key, String value    )                          {        connection.addRequestProperty        (key, value          ); }
//		public Map          getRequestProperties      (                            )                          { return connection.getRequestProperties      (                    ); }
//		public void         setInstanceFollowRedirects(boolean shouldFollow        )                          {        connection.setInstanceFollowRedirects(shouldFollow        ); }
//		public boolean      getInstanceFollowRedirects(                            )                          { return connection.getInstanceFollowRedirects(                    ); }
//		public Permission   getPermission             (                            ) throws IOException       { return connection.getPermission             (                    ); }
	}
	
//### HttpURLConnectionWrapper
	/** HttpURLConnection Υåѡ */
	private class HttpURLConnectionWrapper
		extends HttpURLConnection
	{
		private HttpURLConnection connection;
		
		/** 󥹥󥹤 */
		private HttpURLConnectionWrapper(URL url, HttpURLConnection connection)
		{
			super(url);
			this.connection = connection;
		}
		
		// URLConnection
		public void         connect                   (                            ) throws IOException       {        connection.connect                   (                    ); }
		public URL          getURL                    (                            )                          { return url                                                        ; }
		public int          getContentLength          (                            )                          { return connection.getContentLength          (                    ); }
		public String       getContentType            (                            )                          { return connection.getContentType            (                    ); }
		public String       getContentEncoding        (                            )                          { return connection.getContentEncoding        (                    ); }
		public long         getExpiration             (                            )                          { return connection.getExpiration             (                    ); }
		public long         getDate                   (                            )                          { return connection.getDate                   (                    ); }
		public long         getLastModified           (                            )                          { return connection.getLastModified           (                    ); }
		public String       getHeaderField            (String name                 )                          { return connection.getHeaderField            (name                ); }
		public int          getHeaderFieldInt         (String name, int  Default   )                          { return connection.getHeaderFieldInt         (name, Default       ); }
		public long         getHeaderFieldDate        (String name, long Default   )                          { return connection.getHeaderFieldDate        (name, Default       ); }
		public String       getHeaderFieldKey         (int n                       )                          { return connection.getHeaderFieldKey         (n                   ); }
		public String       getHeaderField            (int n                       )                          { return connection.getHeaderField            (n                   ); }
		public Object       getContent                (                            ) throws IOException       { return connection.getContent                (                    ); }
		public InputStream  getInputStream            (                            ) throws IOException       { return connection.getInputStream            (                    ); }
		public OutputStream getOutputStream           (                            ) throws IOException       { return connection.getOutputStream           (                    ); }
		public String       toString                  (                            )                          { return connection.toString                  (                    ); }
		public void         setDoInput                (boolean doinput             )                          {        connection.setDoInput                (doinput             ); }
		public boolean      getDoInput                (                            )                          { return connection.getDoInput                (                    ); }
		public void         setDoOutput               (boolean dooutput            )                          {        connection.setDoOutput               (dooutput            ); }
		public boolean      getDoOutput               (                            )                          { return connection.getDoOutput               (                    ); }
		public void         setAllowUserInteraction   (boolean allowuserinteraction)                          {        connection.setAllowUserInteraction   (allowuserinteraction); }
		public boolean      getAllowUserInteraction   (                            )                          { return connection.getAllowUserInteraction   (                    ); }
		public void         setUseCaches              (boolean usecaches           )                          {        connection.setUseCaches              (usecaches           ); }
		public boolean      getUseCaches              (                            )                          { return connection.getUseCaches              (                    ); }
		public void         setIfModifiedSince        (long ifmodifiedsince        )                          {        connection.setIfModifiedSince        (ifmodifiedsince     ); }
		public long         getIfModifiedSince        (                            )                          { return connection.getIfModifiedSince        (                    ); }
		public boolean      getDefaultUseCaches       (                            )                          { return connection.getDefaultUseCaches       (                    ); }
		public void         setDefaultUseCaches       (boolean defaultusecaches    )                          {        connection.setDefaultUseCaches       (defaultusecaches    ); }
		public void         setRequestProperty        (String key, String value    )                          {        connection.setRequestProperty        (key, value          ); }
		public String       getRequestProperty        (String key                  )                          { return connection.getRequestProperty        (key                 ); }
//		public void         addRequestProperty        (String key, String value    )                          {        connection.addRequestProperty        (key, value          ); }
//		public Map          getRequestProperties      (                            )                          { return connection.getRequestProperties      (                    ); }
//		public void         setInstanceFollowRedirects(boolean shouldFollow        )                          {        connection.setInstanceFollowRedirects(shouldFollow        ); }
//		public boolean      getInstanceFollowRedirects(                            )                          { return connection.getInstanceFollowRedirects(                    ); }
//		public Permission   getPermission             (                            ) throws IOException       { return connection.getPermission             (                    ); }
		
		// HttpURLConnection
		public void         setRequestMethod          (String method               ) throws ProtocolException {        connection.setRequestMethod          (method              ); }
		public String       getRequestMethod          (                            )                          { return connection.getRequestMethod          (                    ); }
		public int          getResponseCode           (                            ) throws IOException       { return connection.getResponseCode           (                    ); }
		public String       getResponseMessage        (                            ) throws IOException       { return connection.getResponseMessage        (                    ); }
		public void         disconnect                (                            )                          {        connection.disconnect                (                    ); }
		public boolean      usingProxy                (                            )                          { return connection.usingProxy                (                    ); }
//		public InputStream  getErrorStream            (                            )                          { return connection.getErrorStream            (                    ); }
	}
}
