/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.viewer;

import net.hizlab.kagetaka.awt.InnerScrollbar;
import net.hizlab.kagetaka.awt.event.AdjustmentEvent;
import net.hizlab.kagetaka.rendering.PanelScrollbar;

import java.awt.Adjustable;
import java.awt.AWTEvent;
import java.awt.AWTEventMulticaster;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.AdjustmentListener;

/**
 * ѥͥѤΥСǤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.5 $
 */
class ExScrollbar
	extends Component
	implements Adjustable, PanelScrollbar
{
	/** @serial С */
	private Scrollbar scrollbar;
	/** @serial  */
	private int       orientation;
	/** @serial  */
	private boolean   isHorizontal;
	/** @serial ̵ͭ */
	private boolean   enabled = true;
	/** @serial  */
	private int       fixOperator;
	
	/** @serial å */
	private Object    stateLock = new Object();
	/** @serial Ū˺ɽ */
	private boolean   repaintForce;
	
	/** @serial ե꡼ */
	private Image     offScreenImage;
	/** @serial ե꡼ */
	private Dimension offScreenSize;
	/** @serial ե꡼󥰥եå */
	private Graphics  offScreenGraphics;
	
	private transient AdjustmentListener adjustmentListener;
	
	/** 󥹥󥹤 */
	ExScrollbar(int orientation)
	{
		scrollbar = new Scrollbar(orientation);
		scrollbar.setValues(0, 1, 0, 1);
		scrollbar.setEnabled(false);
		
		this.orientation  = orientation;
		this.isHorizontal = (orientation == HORIZONTAL);
		
		setUnitIncrement (1);
		setBlockIncrement(1);
	}
	
//### Override
	/** 侩֤ */
	public Dimension getPreferredSize()
	{
		if (isHorizontal)
			return new Dimension(0, 16);
		else
			return new Dimension(16, 0);
	}
	
	/** 祵֤ */
	public Dimension getMaximumSize()
	{
		return getPreferredSize();
	}
	
	/** ե뤳Ȥ뤫֤ */
	public boolean isFocusTraversable()
	{
		return false;
	}
	
	/** ̵ͭ */
	public synchronized void setEnabled(boolean b)
	{
		if (enabled == b)
			return;
		
		enabled = b;
		super.setEnabled(b);
		scrollbar.setEnabled(b);
		
		repaintForce();
	}
	
	/** ̤򥢥åץǡ */
	public void update(Graphics g)
	{
		paint(g);
	}
	
	/** ᡼ºݤ */
	public void paint(Graphics g)
	{
		Dimension size    = getSize();
		boolean   repaint = false;
		
		synchronized (stateLock) {
			// ѹƤϡե꡼ƺ
			if (offScreenImage == null || !size.equals(offScreenSize)) {
				if (offScreenImage != null) {
					offScreenGraphics.dispose();
					offScreenImage   .flush  ();
				}
				
				if (size.width == 0 || size.height == 0) {
					offScreenImage    = null;
					offScreenSize     = null;
					offScreenGraphics = null;
					return;
				} else {
					offScreenImage    = createImage(size.width, size.height);
					offScreenSize     = size;
					offScreenGraphics = offScreenImage.getGraphics();
				}
				
				scrollbar.setBounds(0, 0, size.width, size.height);
				repaintForce = true;
			}
			
			if (offScreenGraphics != null && repaintForce) {
				repaint      = true;
				repaintForce = false;
			}
		}
		
		if (repaint)
			scrollbar.paint(offScreenGraphics, true);
		
		if (offScreenImage != null)
			g.drawImage(offScreenImage, 0, 0, size.width, size.height, this);
	}
	
	/**
	 * ݡͥȤƥʤ줿ȤΤޤ
	 */
	public void removeNotify()
	{
		synchronized (getTreeLock()) {
			super.removeNotify();
			synchronized (stateLock) {
				if (offScreenImage != null) {
					offScreenGraphics.dispose();
					offScreenImage   .flush  ();
					offScreenGraphics = null;
					offScreenImage    = null;
				}
			}
			scrollbar.dispose();
		}
	}
	
//### Adjustable, PanelScrollbar
	/** ݡͥȤ֤ */
	public Component getComponent()
	{
		return this;
	}
	
	/** ֤ */
	public int getOrientation()
	{
		return orientation;
	}
	
	/** ֤ͤ */
	public int getValue()
	{
		return scrollbar.getValue();
	}
	
	/** ͤꤹ */
	public synchronized void setValue(int newValue)
	{
		if (scrollbar.getValue() == newValue)
			return;
		
		scrollbar.setValue(newValue);
		
		if (!repaintNow())
			repaintForce();
	}
	
	/** Ǿ֤ͤ */
	public int getMinimum()
	{
		return scrollbar.getMinimum();
	}
	
	/** Ǿͤꤹ */
	public void setMinimum(int newMinimum)
	{
		if (scrollbar.getMinimum() == newMinimum)
			return;
		
		int value = scrollbar.getValue();
		if (isHorizontal)
			value -= newMinimum - scrollbar.getMinimum();
		
		fixValue(scrollbar.setValues(value, scrollbar.getVisibleAmount(), newMinimum, scrollbar.getMaximum()));
	}
	
	/** ֤ͤ */
	public int getMaximum()
	{
		return scrollbar.getMaximum();
	}
	
	/** ͤꤹ */
	public synchronized void setMaximum(int newMaximum)
	{
		if (scrollbar.getMaximum() == newMaximum)
			return;
		
		int value = scrollbar.getValue();
		if (isHorizontal)
			value += newMaximum - scrollbar.getMaximum();
		
		fixValue(scrollbar.setValues(value, scrollbar.getVisibleAmount(), scrollbar.getMinimum(), newMaximum));
	}
	
	/** Ļʬ֤ */
	public int getVisibleAmount()
	{
		return scrollbar.getVisibleAmount();
	}
	
	/** Ļʬꤹ */
	public synchronized void setVisibleAmount(int newVisible)
	{
		if (scrollbar.getVisibleAmount() == newVisible)
			return;
		
		int value = scrollbar.getValue();
		if (isHorizontal)
			value -= newVisible - scrollbar.getVisibleAmount();
		
		fixValue(scrollbar.setValues(value, newVisible, scrollbar.getMinimum(), scrollbar.getMaximum()));
	}
	
	/** ˥åȥ󥯥Ȥ֤ */
	public int getUnitIncrement()
	{
		return scrollbar.getUnitIncrement();
	}
	
	/** ˥åȥ󥯥Ȥꤹ */
	public void setUnitIncrement(int v)
	{
		scrollbar.setUnitIncrement(v);
	}
	
	/** ֥å󥯥Ȥ֤ */
	public int getBlockIncrement()
	{
		return scrollbar.getBlockIncrement();
	}
	
	/** ֥å󥯥Ȥꤹ */
	public void setBlockIncrement(int v)
	{
		scrollbar.setBlockIncrement(v);
	}
	
	/** ٥ȥꥹʤϿ */
	public void addAdjustmentListener(AdjustmentListener l)
	{
		enableEvents(AWTEvent.ADJUSTMENT_EVENT_MASK);
		adjustmentListener = AWTEventMulticaster.add(adjustmentListener, l);
	}
	
	/** ٥ȥꥹʤ */
	public void removeAdjustmentListener(AdjustmentListener l)
	{
		adjustmentListener = AWTEventMulticaster.remove(adjustmentListener, l);
	}
	
//### Action
	/**
	 * ΥݡͥȤȯ륳ݡͥȥ٥Ȥޤ
	 * 
	 * @param     e ٥
	 */
	protected void processEvent(AWTEvent e) {
		if (e instanceof AdjustmentEvent) {
			processAdjustmentEvent((AdjustmentEvent)e);
			return;
		}
		super.processEvent(e);
	}
	
	/**
	 * ΥݡͥȤȯ륢㥹ȥȥ٥Ȥ
	 * ϿƤ뤹٤Ƥ {@link AdjustmentListener} 뤳Ȥˤꡢ
	 * 㥹ȥȥ٥Ȥޤ
	 * 
	 * @param     e ٥
	 */
	protected void processAdjustmentEvent(AdjustmentEvent e) {
		if (adjustmentListener != null)
			adjustmentListener.adjustmentValueChanged(e);
	}
	
	/** 㥹ȥȥ٥Ȥȯޤ */
	private void postAdjustmentEvent(int value)
	{
		AdjustmentEvent e = new AdjustmentEvent(this,
		                                        AdjustmentEvent.ADJUSTMENT_VALUE_USER_CHANGED,
		                                        AdjustmentEvent.TRACK,
		                                        value);
		try {
			getToolkit().getSystemEventQueue().postEvent(e);
		} catch (SecurityException ex) {
			processAdjustmentEvent(e);
		}
	}
	
//### private
	/** Ĵ᤹ */
	private void fixValue(boolean repaint)
	{
		boolean b = (scrollbar.getMaximum() > scrollbar.getVisibleAmount());
		if (enabled != b) {
			setEnabled(b);
			return;
		}
		
		if (repaint)
			repaintForce();
	}
	
	/** Ū˺ɽ */
	private void repaintForce()
	{
		synchronized (stateLock) {
			repaintForce = true;
		}
		repaint();
	}
	
	/** ˺ɽ */
	private boolean repaintNow()
	{
		Graphics g;
		synchronized (stateLock) {
			if (offScreenGraphics != null && (g = getGraphics()) != null) {
				try {
					scrollbar.paint(offScreenGraphics, false);
					
					Dimension size = getSize();
					g.drawImage(offScreenImage, 0, 0, size.width, size.height, this);
				} finally {
					g.dispose();
				}
				return true;
			}
		}
		return false;
	}
	
//### Scrollbar
	/** С */
	private class Scrollbar
		extends InnerScrollbar
	{
		/** 󥹥󥹤 */
		private Scrollbar(int orientation)
		{
			super(orientation, ExScrollbar.this);
		}
		
		/**  */
		public void repaint()
		{
			repaintForce();
		}
		
		/** ͤѹ줿 */
		public void changedValue()
		{
			repaintNow();
			postAdjustmentEvent(scrollbar.getValue());
		}
	}
}
