/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.viewer;

import net.hizlab.kagetaka.Reporter;
import net.hizlab.kagetaka.Resource;
import net.hizlab.kagetaka.awt.MessageBox;
import net.hizlab.kagetaka.download.Download;
import net.hizlab.kagetaka.history.History;
import net.hizlab.kagetaka.history.HistoryManager;
import net.hizlab.kagetaka.java2.WheelListener;
import net.hizlab.kagetaka.java2.WheelWrapper;
import net.hizlab.kagetaka.option.OptionListener;
import net.hizlab.kagetaka.option.ViewerController;
import net.hizlab.kagetaka.option.ViewerOption;
import net.hizlab.kagetaka.option.InvalidValueException;
import net.hizlab.kagetaka.rendering.Document;
import net.hizlab.kagetaka.rendering.EngineListener;
import net.hizlab.kagetaka.rendering.HawkContext;
import net.hizlab.kagetaka.rendering.HawkEngine;
import net.hizlab.kagetaka.rendering.HawkPanel;
import net.hizlab.kagetaka.rendering.Option;
import net.hizlab.kagetaka.rendering.PanelListener;
import net.hizlab.kagetaka.rendering.Request;
import net.hizlab.kagetaka.rendering.RequestList;
import net.hizlab.kagetaka.token.FormItem;
import net.hizlab.kagetaka.util.StringUtils;

import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.ImageProducer;
import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * ӥ塼ѤΡ륳ƥȤμǤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.16 $
 */
class ViewerContext
	implements HawkContext,PanelListener,EngineListener,OptionListener
{
	private static final String RESOURCE = "net.hizlab.kagetaka.viewer.Resource";
	
	/** @serial ݡȥץȥ */
	private static Hashtable supportProtocol = new Hashtable();
	{
		StringTokenizer st = new StringTokenizer(Resource.getMessage("config.support.protocol", null), ",; ");
		while (st.hasMoreTokens())
			supportProtocol.put(st.nextToken(), "");
	}
	
	private static int tagIndex;              // ѥǥå
	
	private ViewerOption    option;
	private LocalizedOption localizedOption;
	private Reporter        reporter;
	private HistoryManager  historyManager;
	
	private Connector       connector;
	private HawkEngine      hawkEngine;
	private HawkPanel       hawkPanel;
	private RequestList     requestList;
	private ContextListener contextListener;
	
	private HawkViewer      hawkViewer;
	private MenuItem[][]    appMenu;
	
	// ƥȤξ
	private boolean active;
	private String  tag;
	private String  statusText;
	private long    tabMode;
	
	// ꥯȡ
	private Request viewRequest;              // ɽΥڡΥꥯ
	private Request nextRequest;              // ɤ߹Υꥯ
	private History viewHistory;              // ɽΥڡ
	
	/** 󥹥󥹤 */
	ViewerContext(ViewerOption option)
	{
		this.option          = option;
		this.localizedOption = new LocalizedOption(option);
		this.reporter        = option.getViewerConsole ().getReporter();
		this.historyManager  = option.getHistoryManager();
		this.tag             = String.valueOf(++tagIndex);
		
		connector  = new Connector  (this, option);
		hawkEngine = new HawkEngine (this);
		hawkPanel  = new HawkPanel  (this,
		                             new ExScrollbar(ExScrollbar.HORIZONTAL),
		                             new ExScrollbar(ExScrollbar.VERTICAL  ));
		requestList= new RequestList(    );
		
		hawkEngine.setEngineListener(this);
		hawkPanel .setPanelListener (this);
		
		// ޥۥ
		WheelWrapper wheelWrapper = WheelWrapper.getInstance();
		if (wheelWrapper != null) {
			wheelWrapper.addWheelListener(hawkPanel.getCanvas(), new WheelListener()
				{
					/** ޥۥ뤬ž줿 */
					public void spin(int type, int value, int modifiers)
					{
						if (!ViewerContext.this.option.getPropertyBoolean(ViewerOption.KEY_MOUSE_WHEEL, true))
							return;
						if (ViewerContext.this.option.getPropertyBoolean(ViewerOption.KEY_MOUSE_WHEEL_SWAP, true))
							hawkPanel.scroll(HawkPanel.SCROLL_HORIZONTAL, type, value);
						else
							hawkPanel.scroll(HawkPanel.SCROLL_VERTICAL  , type, value);
					}
				}
			);
		}
		
		// ˥塼ѹ
		initContextMenu();
		
		// ꥹʤϿ
		option.addListener(this);
	}
	
	/** ѥͥΥƥȥ˥塼 */
	private void initContextMenu()
	{
		Vector         items   = null;
		Menu           menu    = null;
		MenuItem       mi      = null;
		ActionListener listener= null;
		
		// ǳ
		items    = hawkPanel.getLinkMenuItems();
		listener = new NewActionListener();
		mi = new MenuItem(getMessage("contextmenu.opennewwindow", null));
		mi.setActionCommand (NewActionListener.WINDOW);
		mi.addActionListener(listener);
		items.insertElementAt(mi, 0);
		mi = new MenuItem(getMessage("contextmenu.opennewtab"   , null));
		mi.setActionCommand (NewActionListener.TAB   );
		mi.addActionListener(listener);
		items.insertElementAt(mi, 1);
		items.insertElementAt(new MenuItem("-"), 2);
		
		// ץꥱǳ˥塼
		initApplicationMenu();
		// ġå
		initToolTip();
	}
	
	/** ѥͥΥƥȥ˥塼ʥץꥱˤ */
	private void initApplicationMenu()
	{
		Menu           menu    = null;
		Vector         items   = null;
		ActionListener listener= null;
		
		// ץꥱǳ˥塼
		ViewerOption.Application[] apps = option.getApplications();
		
		if (apps == null || apps.length == 0) {
			if (appMenu == null)
				return;
			// Ͽ
			items = hawkPanel.getMainMenuItems();
			items.removeElement(appMenu[0][0]);
			items.removeElement(appMenu[0][1]);
			items = hawkPanel.getLinkMenuItems();
			items.removeElement(appMenu[1][0]);
			items.removeElement(appMenu[1][1]);
			appMenu = null;
			return;
		}
		
		String label = getMessage("contextmenu.openapp", null);
		if (appMenu == null)
			appMenu = new MenuItem[2][2];
		
		for (int i = 0; i < 2; i++) {
			switch (i) {
			case 0:
				listener = new ApplicationActionListener(ApplicationActionListener.MAIN, option);
				items    = hawkPanel.getMainMenuItems();
				break;
			case 1:
				listener = new ApplicationActionListener(ApplicationActionListener.LINK, option);
				items    = hawkPanel.getLinkMenuItems();
				break;
			}
			
			if (appMenu[i][0] == null) {
				items.addElement(appMenu[i][0] = new MenuItem("-"  ));
				items.addElement(appMenu[i][1] = new Menu    (label));
			}
			menu = (Menu)appMenu[i][1];
			menu.removeAll();
			
			for (int j = 0; j < apps.length; j++)
				createApplicationMenu(apps[j], menu, listener);
		}
	}
	
	/** ġåפ */
	private void initToolTip()
	{
		hawkPanel.setToolTipOption(option.getPropertyBoolean(ViewerOption.KEY_TOOLTIP_BROWSER_USE  , true),
		                           option.getPropertyInteger(ViewerOption.KEY_TOOLTIP_BROWSER_DELAY, 0   ),
		                           option.getPropertyInteger(ViewerOption.KEY_TOOLTIP_BROWSER_TERM , 0   ));
	}
	
	/** ץꥱ˥塼 */
	private void createApplicationMenu(ViewerOption.Application app,
	                                   Menu parent,
	                                   ActionListener listener)
	{
		MenuItem mi = new MenuItem(app.getName());
		
		mi.setActionCommand (app.getId());
		mi.addActionListener(listener   );
		
		parent.add(mi);
	}
	
	/** ѥ֤ͥ */
	HawkPanel attachViewer(HawkViewer hawkViewer)
	{
		this.hawkViewer = hawkViewer;
		connector.setOwner(hawkViewer);
		return hawkPanel;
	}
	
	/** ѥͥΥХ֤ */
	Component getCanvas()
	{
		return hawkPanel.getCanvas();
	}
	
	/** ƥȥꥹʤɲ */
	void addContextListener(ContextListener l)
	{
		this.contextListener = ContextEventMulticaster.add(contextListener, l);
	}
	
	/** ƥȥꥹʤ */
	void removeContextListener(ContextListener l)
	{
		this.contextListener = ContextEventMulticaster.remove(contextListener, l);
	}
	
	/** ꥯȤ */
	void load(Request request)
	{
		hawkEngine.stop();
		// ɤ߹ߤξǡPOST ǡ¸ߤϳǧ
		if (request.getDocument() != null &&    // ɤ߹߻ˤ Document ͭ
		    request.postData      != null &&
		    MessageBox.show(hawkViewer,
		                    getMessage("message.dorepost.text" , null),
		                    getMessage("message.dorepost.title", null),
		                    MessageBox.BUTTON_OKCANCEL | MessageBox.ICON_QUESTION) != MessageBox.BUTTON_OK)
			return;
		
		request.setMoved(false);
		
		synchronized (this) {
			// ƱڡΥե󥹤ˤưǺɤ߹ߤʤ
			if (viewRequest != null                      &&
			    viewRequest != request                   &&        // ɤ߹ߤǤϤʤ
			    request.useCache >= Request.CACHE_NORMAL &&        // åѲ
			    StringUtils.sameFile(viewRequest.getDocument().url, request.url)) {
				request.copyFrom(viewRequest);
				changeHistory(request);
				movePosition (true   );
				setStatus    (null   );
				return;
			}
//System.out.println("URL=["+url+"]");
			nextRequest = request;
			contextListener.documentLoading(this);
		}
		
		hawkPanel.setCursor(Cursor.WAIT_CURSOR);
		hawkEngine.load(request);
	}
	
	/** ƥ֤ */
	void setActived(boolean active)
	{
		this.active = active;
		
		if (active) {
			// ӥ塼ɽ򹹿
			Request request = viewRequest;
			
			// ꥹʡư
			contextListener.activeChanged(this);
			if (request != null) {
				Document document = request.getDocument();
				setViewerTitle  (document.getTitle    ());
				setViewerAddress(document.url.toString());
			} else {
				setViewerTitle  (null);
				setViewerAddress(null);
			}
			setViewerStatus(statusText);
		}
	}
	
	/** ꥯȤ */
	Request getRequest()
	{
		return viewRequest;
	}
	
	/** 椫 */
	boolean isLoading()
	{
		return (nextRequest != null);
	}
	
	/** ȥ */
	String getTitle()
	{
		if (viewRequest == null)
			return "";
		
		String title = viewRequest.getDocument().getTitle();
		
		return (title != null ? title : "");
	}
	
	/** URL  */
	URL getURL()
	{
		return (viewRequest != null ? viewRequest.url : null);
	}
	
	/**  */
	String getTag()
	{
		return tag;
	}
	
	/** ɤ߹椫ɤ */
	void repostLoadEvent()
	{
		//### BUGS 
		synchronized (this) {
			if (nextRequest != null)
				contextListener.documentLoading(this);
			else
				contextListener.documentLoaded (this);
		}
	}
	
	/** ⡼ɤ */
	long getTabMode()
	{
		return tabMode;
	}
	
	/** ⡼ɤɲ */
	void addTabMode(long mode)
	{
		tabMode |= mode;
	}
	
	/** ⡼ɤ */
	void removeTabMode(long mode)
	{
		tabMode &= ~mode;
	}
	
	/**  */
	void back(int num)
	{
		Request request = viewRequest;
		Request r;
		while (num-- > 0 && (r = request.getPrevious()) != null)
			request = r;
		
		if (request != viewRequest)
			load(request);
	}
	
	/** ؿʤ */
	void forward(int num)
	{
		Request request = viewRequest;
		Request r;
		while (num-- > 0 && (r = request.getNext()) != null)
			request = r;
		
		if (request != viewRequest)
			load(request);
	}
	
	/** ˴ */
	void dispose()
	{
		hawkEngine.dispose();
	}
	
	//### HawkContext
	/** 륦ɥѤα륳ƥȤ */
	public Reporter getReporter()
	{
		return reporter;
	}
	
	/** ġ륭åȤ֤ */
	public Toolkit getToolkit()
	{
		return hawkPanel.getToolkit();
	}
	
	/** ꤷꥯȤɽ */
	public void openHawk(Request request)
	{
		String protocol = request.url.getProtocol();
		
		if (!supportProtocol.containsKey(protocol)) {
			MessageBox.show(hawkViewer,
			                getMessage("message.unknownprotocol.text" , new String[]{protocol}),
			                getMessage("message.unknownprotocol.title", null),
			                MessageBox.BUTTON_OK | MessageBox.ICON_EXCLAMATION);
			return;
		}
		
		hawkViewer.open(request);
	}
	
	/** ꤵ줿 URL Υ֤ͥ */
	public URLConnection getURLConnection(Request request)
		throws IOException, InterruptedException
	{
		return connector.getURLConnection(request);
	}
	
	/**  ꤵ줿ɥȤ */
	public void download(Document document, URLConnection connection)
	{
		Download.show(hawkViewer, option, connection, null);
	}
	
	/** ᡼֤ޤ */
	public Image createImage(ImageProducer producer)
	{
		return hawkPanel.createImage(producer);
	}
	
	/** ᡼֤ޤ */
	public Image createImage(int width, int height)
	{
		return hawkPanel.createImage(width, height);
	}
	
	/** ǥȥå֤ */
	public MediaTracker getMediaTracker()
	{
		return hawkPanel.getMediaTracker();
	}
	
	/** ꡼󥤥᡼ɲ */
	public void addImage(Image image, int width, int height)
	{
		hawkPanel.addImage(image, width, height);
		synchronized (this) {
			movePosition(false);
		}
	}
	
	/** եॢƥɲ */
	public void addFormItem(FormItem item)
	{
		hawkPanel.addFormItem(item);
	}
	
	/** ѥ򥻥åȥå */
	public void setupCanvas(Request request)
	{
		String title = request.getDocument().getTitle();
		
		viewHistory.setTitle(title);
		setViewerTitle      (title);
		
		hawkPanel.setupCanvas(request);
	}
	
	/** ѥϰϤꤷƺɽ */
	public void repaint(int x, int y, int width, int height)
	{
		hawkPanel.repaintCanvas(x, y, width, height);
	}
	
	/** ơå */
	public void setStatus(String status)
	{
		statusText = status;
		
		setViewerStatus(status);
	}
	
	/** Ūʥơå */
	public void setTemporaryStatus(String status)
	{
		setViewerStatus((status != null ? status: statusText));
	}
	
	/** ӥ塼ݡȤθߤΥ֤ */
	public Dimension getViewportSize()
	{
		return hawkPanel.getViewportSize();
	}
	
	/** ץ֤ */
	public Option getOption()
	{
		return localizedOption;
	}
	
	/** ꤷ URL ˬ䤷Ȥ뤫ɤ֤ */
	public boolean haveEverVisited(URL url)
	{
		return historyManager.contains(url.toString());
	}
	
	/** å */
	public void setCookie(String value, URL url)
	{
		connector.setCookie(value, url);
	}
	
	//### PanelListener
	/**  */
	public void back()
	{
		back(1);
	}
	
	/** 뤫ɤ */
	public boolean canBack()
	{
		return (viewRequest != null && viewRequest.getPrevious() != null);
	}
	
	/** ؿʤ */
	public void forward()
	{
		forward(1);
	}
	
	/** ؿʤ뤫ɤ */
	public boolean canForward()
	{
		return (viewRequest != null && viewRequest.getNext() != null);
	}
	
	/** ɹ */
	public void reload(boolean force)
	{
		if (viewRequest != null)
			load(viewRequest);
	}
	
	/** ɹǤ뤫ɤ */
	public boolean canReload()
	{
		return (viewRequest != null);
	}
	
	/**  */
	public void stop()
	{
		hawkEngine.stop();
	}
	
	/** ߤǤ뤫ɤ */
	public boolean canStop()
	{
		return hawkEngine.isLoading();
	}
	
	//### EngineListener
	/** ³Ϥ˸ƤӽФ */
	public void connecting(Request request)
	{
/*
		if (nextRequest != request)
			return;
*/
	}
	
	/** ³λȸƤӽФ */
	public synchronized void connected(Request request, boolean noerror)
	{
/*
		if (nextRequest != request)
			return;
*/
	}
	
	/** 褬åפ줿Ȥ˸ƤӽФ */
	public synchronized void renderingSkipped(Request request)
	{
		if (nextRequest != request)
			return;
		
		nextRequest = null;
		hawkPanel.setCursor(Cursor.DEFAULT_CURSOR);
		contextListener.documentLoaded(this);
	}
	
	/** 褬Ϥ˸ƤӽФ */
	public synchronized void renderingStarted(Request request)
	{
		if (nextRequest != request)
			return;
		
		changeHistory(request);
		hawkPanel.requestFocus();
	}
	
	/** 褬λȸƤӽФ */
	public synchronized void renderingStopped(Request request, boolean noerror)
	{
		if (nextRequest != request)
			return;
		
		renderingSkipped(request);
		movePosition(true);
	}
	
	/** ꤷ URL ɽ */
	public synchronized void refresh(Request baseRequest, Request newRequest)
	{
		if (viewRequest != baseRequest)
			return;
		
		load(newRequest);
	}
	
	//### OptionListener
	/** ͤѹȤ */
	public void propertyChange(ViewerOption option, String key, Object oldValue, Object newValue)
		throws InvalidValueException
	{
	}
	
	/** ͤѹꡢѹꤵ줿 */
	public void propertiesChanged(ViewerOption option, ViewerController c, Enumeration list)
	{
		String key;
		while (list.hasMoreElements()) {
			key = (String)list.nextElement();
			
			// ץꥱ
			if (key.startsWith(ViewerOption.KEY_APP)) {
				initApplicationMenu();
				return;
			}
			
			// ʡġåפ
			if (key.compareTo(ViewerOption.KEY_TOOLTIP_BROWSER_USE  ) == 0 ||
			    key.compareTo(ViewerOption.KEY_TOOLTIP_BROWSER_DELAY) == 0 ||
			    key.compareTo(ViewerOption.KEY_TOOLTIP_BROWSER_TERM ) == 0)
				initToolTip();
		}
	}
	
	//### private
	/** ѹץå */
	private void changeHistory(Request request)
	{
		// ߤɽ֤¸
		if (viewRequest != null)
			viewRequest.setPosition(hawkPanel.getScrollPosition());
		
		// ξ
		if (!request.isMember())
			requestList.append(viewRequest, request);
		
		viewRequest = request;
		viewHistory = historyManager.visited(request.url.toString());
		
		setViewerAddress(request.getDocument().url.toString());
	}
	
	/** ݥưץå */
	private void movePosition(boolean force)
	{
		if (viewRequest.isMoved())
			return;
		
		Point position;
		
		// 桼ư֤
		if ((position = viewRequest.getPosition()) != null) {
			// 桼ˤꥯ뤵Ƥʤ
			Point scpos = hawkPanel.getScrollPosition();
			if ((force || (scpos.x == 0 && scpos.y == 0)) &&
			    !hawkPanel.setScrollPosition(position.x, position.y, false))
				return;
		} else
		
		// # ե󥹤ΰ֤
		if (!hawkPanel.setScrollPosition(viewRequest.getDocument().url.getRef(), force))
			return;
		
		viewRequest.setMoved(true);
	}
	
	/** ꥽ʸ */
	private String getMessage(String key, String[] args)
	{
		return Resource.getMessage(RESOURCE, key, args);
	}
	
//### Viewer Ф
	/** ȥ */
	private void setViewerTitle(String title)
	{
		contextListener.titleChanged(this, title);
	}
	
	/** ɥ쥹 */
	private void setViewerAddress(String address)
	{
		contextListener.addressChanged(this, address);
		contextListener.historyChanged(this);
	}
	
	/** ơС */
	private void setViewerStatus(String status)
	{
		contextListener.statusChanged(this, status);
	}
	
//### NewActionListener
	/** ǳѤΥꥹ */
	private class NewActionListener
		implements ActionListener
	{
		private static final String WINDOW = "open.newwindow";
		private static final String TAB    = "open.newtab";
		
		/** ꥹʤ */
		private NewActionListener()
		{
		}
		
		/** ϥɥ */
		public void actionPerformed(ActionEvent e)
		{
			String command = e.getActionCommand();
			
			URL url = hawkPanel.getSelectedLink();
			if (url == null)
				return;
			
			URL referer = (viewRequest != null ? viewRequest.getDocument().url : null);
			
			if (WINDOW.compareTo(command) == 0) {
				openHawk(new Request(url, null, referer, null, true, false, Request.CACHE_NORMAL));
			} else
			if (TAB   .compareTo(command) == 0) {
				openHawk(new Request(url, null, referer, null, false, true, Request.CACHE_NORMAL));
			}
		}
	}
	
//### ApplicationActionListener
	/** ץꥱ󤫤鳫ѤΥꥹ */
	private class ApplicationActionListener
		implements ActionListener
	{
		private static final int MAIN = 0;
		private static final int LINK = 1;
		
		private int          type;
		private ViewerOption option;
		
		/** ꥹʤ */
		private ApplicationActionListener(int type, ViewerOption option)
		{
			this.type   = type;
			this.option = option;
		}
		
		/** ϥɥ */
		public void actionPerformed(ActionEvent e)
		{
			String command = e.getActionCommand();
			ViewerOption.Application[] apps = option.getApplications();
			if (apps == null)
				return;
			
			for (int i = 0; i < apps.length; i++) {
				ViewerOption.Application app = apps[i];
				if (command.compareTo(app.getId()) != 0)
					continue;
				
				URL url = (type == MAIN ? (viewRequest != null ? viewRequest.url : null)
				                        : hawkPanel.getSelectedLink());
				if (url == null)
					return;
				
				StringBuffer path = new StringBuffer();
				
				try {
					app.exec(url, path);
				} catch (IOException ex) {
					MessageBox.show(hawkViewer,
					                getMessage("message.appnotopen.text" , new String[]{path.toString(), url.toString(), ex.toString()}),
					                getMessage("message.appnotopen.title", null),
					                MessageBox.BUTTON_OK | MessageBox.ICON_EXCLAMATION);
				}
			}
		}
	}
}
